_find_user_query = \
'''query search_chars( $char_email: String!){
  users{
    search(query: $char_email) {
      id
      email
    }
  }
}'''

_user_single_query = \
'''query get_user($uid: Int!{
  users{
    single(id: $uid) {
      id
      email
      createdAt
      isActive
    }
  }
}'''

_get_group_list_query = \
'''query {
  groups{
    list(filter:"", orderBy:""){
      id
      name
      isSystem
    }
  }
}'''

_create_group_mutation = \
'''mutation create_group($group_name:String!){
  groups{
    create(name: $group_name){
      group{
       id
       name
      }
      responseResult {
        message
        succeeded
        errorCode
        slug
      }
    }
  }
}'''

_update_user_mutation = \
'''mutation update_user_groups(
  $group_list:[Int]!
  $name:String!
  $uid:Int!
	){
  users{
    update(
      id: $uid
      name: $name
      groups: $group_list
    ){
      responseResult {
        message
        succeeded
        errorCode
        slug
      }
    }
  }
}'''

_create_user_mutation = \
'''mutation create_user(
  $group_list: [Int]!
  $name: String!
  $pass: String!
  $email: String!
) {
  users {
    create(
      email: $email
      name: $name
      passwordRaw: $pass
      groups: $group_list
      providerKey: "local"
      sendWelcomeEmail: false
      mustChangePassword: false
    ) {
      responseResult {
        errorCode
        succeeded
        message
        slug
      }
    }
  }
}
'''
_user_password_mutation = \
'''mutation update_user(
  $password: String!
  $uid: Int!
) {
  users {
    update(
      id: $uid
      newPassword: $password
    ) {
      responseResult {
        errorCode
        succeeded
        message
        slug
      }
    }
  }
}
'''

_deactivate_user_mutation = \
'''mutation deactivate_user(
  $uid: Int!
) {
  users {
    deactivate(
        id: $uid
    ) {
      responseResult {
        errorCode
        succeeded
        message
        slug
      }
    }
  }
}
'''

_activate_user_mutation = \
'''mutation deactivate_user(
  $uid: Int!
) {
  users {
    activate(
        id: $uid
    ) {
      responseResult {
        errorCode
        succeeded
        message
        slug
      }
    }
  }
}
'''

_find_pages_query = \
'''query find_pages($search_str:String!, $locale:String!){
  pages{
    search(query:$search_str,locale: $locale){
      results{
        id
        title
        description
        path
      }
      suggestions
    }
  }
}
'''



from django.contrib.auth.models import User
from django.db import models


class WikiJs(models.Model):
    user = models.OneToOneField(User,
                                primary_key=True,
                                on_delete=models.CASCADE,
                                related_name='wikijs')

    uid = models.BigIntegerField()

    def __str__(self):
        return "WikiJS User - {}".format(self.user.username)

    class Meta:
        permissions = (
            ("access_wikijs", u"Can access the WikiJS service"),
        )



# Generated by Django 2.2.12 on 2020-06-11 10:28

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0011_update_proxy_permissions'),
    ]

    operations = [
        migrations.CreateModel(
            name='WikiJsUser',
            fields=[
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='wikijs', serialize=False, to=settings.AUTH_USER_MODEL)),
                ('uid', models.BigIntegerField()),
            ],
            options={
                'permissions': (('access_wikijs', 'Can access the WikiJS service'),),
            },
        ),
    ]



# Cog used by https://github.com/pvyParts/allianceauth-discordbot

# Cog Stuff
from discord.ext import commands
from discord.embeds import Embed
from discord.colour import Color

# AA Contexts
from aadiscordbot.cogs.utils.decorators import sender_has_perm
from wikijs.manager import WikiJSManager
from django.conf import settings

import re

import logging
logger = logging.getLogger('aadiscordbot.cogs.wikijs')


class Wikijs(commands.Cog):
    """
    WikiJS relevant cogs for AADiscordbot
    Currently onlt implements minimal search
    """
    def __init__(self, bot):
        self.bot = bot

    @commands.command(pass_context=True)
    @sender_has_perm('wikijs.access_wikijs')
    async def wiki(self, ctx):
        """
        Returns the top Wiki Article search result for a string
        """
        logger.debug("WikiJS Cog: !wikijs received")
        await ctx.channel.trigger_typing()
        await ctx.message.add_reaction(chr(0x231B))

        search_string = ctx.message.content[6:]
        if search_string == "":
            embed = Embed(title="WikiJS")
            embed.colour = Color.blue()
            embed.add_field(
                name="Wiki Link",
                value=f"{settings.WIKIJS_URL}"
            )
            await ctx.message.clear_reaction(chr(0x231B))
            return await ctx.reply(embed=embed, mention_author=False)

        try:
            pagesearchresponse = WikiJSManager().search_for_page(search_string)
            logger.debug(f"WikiJS Cog: page search response {pagesearchresponse}")
        except Exception as e:
            logger.error(e)

        embed = Embed(title=f"WikiJS Search: {search_string}")

        for result in pagesearchresponse["data"]["pages"]["search"]["results"]:
            logger.debug(f"WikiJS Cog: single page {result}")
            title = result["title"]
            path = result["path"]
            embed.add_field(
                name=title,
                value=f"{settings.WIKIJS_URL}{path}"
            )

        await ctx.message.clear_reaction(chr(0x231B))
        return await ctx.reply(embed=embed, mention_author=False)


def setup(bot):
    bot.add_cog(Wikijs(bot))



import logging

from django.template.loader import render_to_string
from django.conf import settings

from allianceauth import hooks
from allianceauth.services.hooks import ServicesHook
from django.core.exceptions import ObjectDoesNotExist
from django.utils.crypto import get_random_string
from wikijs.app_settings import WIKIJS_AADISCORDBOT_INTEGRATION

from .urls import urlpatterns
from .manager import WikiJSManager
from .tasks import WikiJSTasks
from .models import WikiJs
logger = logging.getLogger(__name__)


class WikiJSService(ServicesHook):
    def __init__(self):
        ServicesHook.__init__(self)
        self.urlpatterns = urlpatterns
        self.name = 'Wiki JS'
        self.service_url = settings.WIKIJS_URL
        self.access_perm = 'wikijs.access_wikijs'
        self.name_format = '{character_name}'

    @property
    def title(self):
        return self.name

    def delete_user(self, user):
        logger.debug('Deleting user %s %s account' % (user, self.name))
        return WikiJSManager().deactivate_user(user)

    def validate_user(self, user):
        logger.debug('Validating user %s %s account' % (user, self.name))
        if self.user_has_account(user) and not self.service_active_for_user(user):
            WikiJSManager().deactivate_user(user)

    def update_groups(self, user):
        logger.debug('Updating %s groups for %s' % (self.name, user))
        if self.user_has_account(user):
            return WikiJSTasks.update_member.delay(user.pk)
        return False

    def update_all_groups(self):
        logger.debug('Update all %s groups called' % self.name)
        for u in WikiJs.objects.all():
            self.validate_user(u.user)
            WikiJSTasks.update_member.delay(u.user_id)

    def service_active_for_user(self, user):
        return user.has_perm(self.access_perm)

    def render_services_ctrl(self, request):
        urls = self.Urls()
        urls.auth_activate = 'wikijs:activate'
        urls.auth_deactivate = 'wikijs:deactivate'
        urls.auth_reset_password = 'wikijs:reset_password'
        urls.auth_set_password = 'wikijs:set_password'
        return render_to_string(self.service_ctrl_template, {
            'service_name': self.title,
            'urls': urls,
            'service_url': self.service_url,
            'username': request.user.email if self.user_has_account(request.user) else ''
        }, request=request)

    def user_has_account(self, user):
        try:
            user.wikijs
        except ObjectDoesNotExist:
            return False
        else:
            return True


@hooks.register('services_hook')
def register_service():
    return WikiJSService()


@hooks.register('discord_cogs_hook')
def register_cogs():
    if WIKIJS_AADISCORDBOT_INTEGRATION is True:
        return ["wikijs.cogs.wikijs"]
    else:
        return [""]



import os
from setuptools import find_packages, setup
from wikijs import VERSION

install_requires = [
    'graphqlclient',
    'allianceauth'
]

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='allianceauth-wiki-js',
    version=VERSION,
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    license='MIT',
    description='Alliance Auth Service module for Wiki JS',
    long_description=long_description,
    url="https://github.com/pvyParts/allianceauth-wiki-js",
    long_description_content_type='text/markdown',
    author='AaronKable',
    author_email='aaronkable@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',        
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



import logging

from django.conf import settings
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from allianceauth.notifications import notify
from celery import shared_task
from allianceauth.services.tasks import QueueOnce
import requests

from .manager import WikiJSManager
from .models import WikiJs

logger = logging.getLogger(__name__)

class WikiJSTasks:
    @staticmethod
    @shared_task(bind=True, name='wikijs.update_member', base=QueueOnce)
    def update_member(self, user_id):
        u = User.objects.get(id=user_id)
        try:
            u.wikijs
        except ObjectDoesNotExist:
            return "User has no Wiki.JS account"
        else:
            return f"Updated Wiki.JS for {u.username}, Result:{WikiJSManager().update_user(u)}"
    
    @staticmethod
    @shared_task(bind=True, name='wikijs.update_all_members', base=QueueOnce)
    def update_all_members(self):
        for u in WikiJs.objects.all():
            WikiJSTasks.update_member.delay(u.pk)



import json
from allianceauth.services.hooks import get_extension_logger
import re
from hashlib import md5

import requests
from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist
from django.utils.crypto import get_random_string
from graphqlclient import GraphQLClient

from allianceauth.services.hooks import NameFormatter

from .models import WikiJs
from .queries import (
    _activate_user_mutation,
    _create_group_mutation,
    _create_user_mutation,
    _deactivate_user_mutation,
    _find_user_query,
    _get_group_list_query,
    _update_user_mutation,
    _user_password_mutation,
    _user_single_query,
    _find_pages_query,
)

logger = get_extension_logger(__name__)

GROUP_CACHE_MAX_AGE = getattr(settings, 'WIKIJS_GROUP_CACHE_MAX_AGE', 2 * 60 * 60)  # default 2 hours

class WikiJSManager:
    _client = None
    def __init__(self):
        try:
            self.client
        except Exception as e:
            print(e)

    @property
    def client(self):
        if not self._client:
            self._client = GraphQLClient("{}/graphql".format(settings.WIKIJS_URL))
            self._client.inject_token("Bearer {}".format(settings.WIKIJS_API_KEY))
        return self._client

    ### Groups! ****************************************************************************************************

    def _get_groups(self):
        data = json.loads(self.client.execute(_get_group_list_query)).get("data",{}).get("groups",{}).get("list",[])
        return data

    def _create_group(self, name):
        data = json.loads(self.client.execute(_create_group_mutation, variables={"group_name":name}))
        try:
            if not data["data"]["groups"]["create"]["responseResult"]['succeeded']:
                logger.error("WikiJs unable to create group. {}".format(data["data"]["groups"]["create"]["responseResult"]["message"]))
                return None
        except:
            logger.error("API returned invalid response when creating group {}".format(data), exc_info=1)
        return data.get("data",{}).get("groups").get("create").get("group")


    def __group_name_to_id(self, name):
        name = WikiJSManager._sanitize_groupname(name)

        def get_or_create_group():
            groups = self._get_groups()
            if groups is not None:
                for g in groups:
                    if g['name'].lower() == name.lower():
                        return g['id']
            return self._create_group(name)['id']

        return cache.get_or_set(WikiJSManager._generate_cache_group_name_key(name), get_or_create_group,
                                GROUP_CACHE_MAX_AGE)

    def __group_id_to_name(self, g_id):
        def get_group_name():
            groups = self._get_groups()
            for g in groups:
                if g['id'] == g_id:
                    return g['name']
            raise KeyError("Group ID %s not found on Wiki.JS" % g_id)

        return cache.get_or_set(WikiJSManager._generate_cache_group_id_key(g_id), get_group_name,
                                GROUP_CACHE_MAX_AGE)
    
    def __generate_group_list(self, names):
        group_list = []
        for name in names:
            group_list.append(self.__group_name_to_id(name))
        return group_list

    ### Pages ********************************

    def __find_pages(self, search_string, locale):
        logger.debug(f"Hitting API looking for page {search_string}")
        data = json.loads(self.client.execute(_find_pages_query, variables={"search_str":search_string, "locale": locale}))
        return data
    
    ### Users *****************************************************************************************************
    def __find_user(self, email):
        logger.debug("Hitting API looking for: {}".format(email[:10]))
        data = json.loads(self.client.execute(_find_user_query, variables={"char_email":email.lower()}))
        logger.debug("API returned: {}".format(data))
        users = data.get("data", {}).get("users", {}).get("search", [])
        if users is None:
            return False
        for user in users:
            if user.get("email", "").lower() == email.lower():
                return user.get("id")
        return False

    def __create_user(self, user, password=False):
        from .auth_hooks import WikiJSService

        name = NameFormatter(WikiJSService(), user).format_name()

        if not password:
            password = get_random_string(15)

        groups = [WikiJSManager._sanitize_groupname(user.profile.state.name)]
        for g in user.groups.all():
            groups.append(WikiJSManager._sanitize_groupname(str(g)))
        
        group_list = self.__generate_group_list(groups)
        data = json.loads(self.client.execute(_create_user_mutation, 
                            variables={
                                "group_list":group_list,
                                "email":user.email.lower(),
                                "name":name,
                                "pass":password}))
        logger.debug("API returned: {}".format(data))
        if data["data"]["users"]["create"]["responseResult"]["succeeded"]:
            uid = self.__find_user(user.lower())
            if uid:
                WikiJs.objects.update_or_create(user=user, uid=uid)
                return uid
        else:
            logger.error("WikiJs unable to Create User. {}".format(data["data"]["users"]["create"]["responseResult"]["message"]))
        return False

    def __deactivate_user(self, uid):
        data = json.loads(self.client.execute(_deactivate_user_mutation, variables={"uid":uid}))
        result = data["data"]["users"]["deactivate"]["responseResult"]["succeeded"]
        if not result:
            logger.error("WikiJs unable to deactivate User. {}".format(data["data"]["users"]["deactivate"]["responseResult"]["message"]))
        else:
            WikiJs.objects.filter(uid=uid).delete()
        return result

    def __activate_user(self, uid):
        data = json.loads(self.client.execute(_activate_user_mutation, variables={"uid":uid}))
        result = data["data"]["users"]["activate"]["responseResult"]["succeeded"]
        if not result:
            logger.error("WikiJs unable to activate User. {}".format(data["data"]["users"]["activate"]["responseResult"]["message"]))
        return result

    def _update_password(self, uid, password):
        data = json.loads(self.client.execute(_user_password_mutation, 
                            variables={
                                "uid":uid,
                                "password":password
                                }))
        result = data["data"]["users"]["update"]["responseResult"]["succeeded"]
        if not result:
            logger.error("WikiJs unable to update password for User. {}".format(data["data"]["users"]["update"]["responseResult"]["message"]))
        return result

    def _update_user(self, user):
        from .auth_hooks import WikiJSService

        groups = [WikiJSManager._sanitize_groupname(user.profile.state.name)]
        for g in user.groups.all():
            groups.append(WikiJSManager._sanitize_groupname(str(g)))
        group_list = self.__generate_group_list(groups)
        name = NameFormatter(WikiJSService(), user).format_name()

        data = json.loads(self.client.execute(_update_user_mutation, 
                            variables={
                                "uid":user.wikijs.uid,
                                "name":name,
                                "group_list":group_list
                                }))
        result = data["data"]["users"]["update"]["responseResult"]["succeeded"]
        if not result:
            logger.error("WikiJs unable to update User. {}".format(data["data"]["users"]["update"]["responseResult"]["message"]))
        return result


    #Statics ******************************************************************************************************
    @staticmethod
    def _generate_cache_group_name_key(name):
        return 'WIKIJS_GROUP_NAME__%s' % md5(name.encode('utf-8')).hexdigest()

    @staticmethod
    def _generate_cache_group_id_key(g_id):
        return 'WIKIJS_GROUP_ID__%s' % g_id

    @staticmethod
    def _sanitize_name(name):
        name = name.replace(' ', '_')
        name = name.replace("'", '')
        name = name.lstrip(' _')
        name = name[:20]
        name = name.rstrip(' _')
        return name

    @staticmethod
    def _sanitize_groupname(name):
        name = re.sub('[^\w]', '', name)
        name = WikiJSManager._sanitize_name(name)
        if len(name) < 3:
            name = "Group " + name
        return name

    @staticmethod
    def user_has_account(user):
        try:
            user.wikijs
        except ObjectDoesNotExist:
            return False
        else:
            return True

    ### Methods **************************************************************************************************
    def update_user(self, user):
        return self._update_user(user)


    def activate_user(self, user):
        #search
        try:
            uid = self.__find_user(user.email.lower())
            #create
            if not uid:
                logger.info("Creating new user for {}".format(user.username))
                uid = self.__create_user(user)
            else:
                logger.info("reactivating disabled account for {}".format(user.username))
                self.__activate_user(uid)
                WikiJs.objects.update_or_create(user=user, uid=uid)
                self.update_user(user)

            #password
            password = get_random_string(15)
            self._update_password(uid, password)

            #return
            return password
        except Exception as e:
            logger.error(e,exc_info=1)
            return False

    def deactivate_user(self, user):

        try:
            result = self.__deactivate_user(user.wikijs.uid)
        except AttributeError: #no wikijs model
            return True

        if result:
            try:
                user.wikijs.delete()
            except:
                pass
        return result

    def search_for_page(self, search_string: str, locale: str = 'en') -> dict:
        """
        Returns Search results for a WikiJS page search
        """
        try: 
            result = self.__find_pages(search_string, locale)
        except Exception as e:
            logger.error(e)

        return result
